// basic-kernel.h
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Author: allauzen@cs.nyu.edu (Cyril Allauzen)
//
// \file
// Simple and modular cached kernel representation.

#ifndef KERNEL_LIB_BASIC_KERNEL_H__
#define KERNEL_LIB_BASIC_KERNEL_H__

#include "kernel/lib/kernel.h"
#include "kernel/lib/cache.h"

namespace kernel {


template <class F, class C>
class BasicKernelImpl : public KernelImpl<typename F::Object> {
 public:
  typedef F Function;
  typedef typename Function::Object Object;
  typedef C Cache;

  using KernelImpl<Object>::Data;
  using KernelImpl<Object>::SetData;

  BasicKernelImpl(const F *function, C* cache, bool normalize)
  : normalize_(normalize) {
    KernelImpl<Object>::SetType("basic");
    cache_ = cache;
    function_ = function;
    uint64 props = normalize ? function->Properties() | kNormalized
                   : function->Properties();
    KernelImpl<Object>::SetProperties(props);
  }

  ~BasicKernelImpl() {
    VLOG(1) << "~BasicKernelImpl[" << KernelImpl<Object>::RefCount() << "]";
    delete cache_;
    delete function_;
  }

  double Eval(size_t i, size_t j) { return Eval(i, j, normalize_); }

  double Eval(size_t i, size_t j, bool normalize) {
    if (normalize) {
      if (i == j) return 1.0F;
      double nu = sqrt(Eval(i, i, false) * Eval(j, j, false));
      if (nu == 0.0)
        LOG(ERROR) << "BasicKernel::Eval(" << i << "," << j
                   << "): cannot normalize kernel when"
                   << "k(x,x) = 0 for some x";
      return Eval(i, j, false) / nu;
    }
    uint32 type = cache_->CacheType();
    if ((type & kCacheValue) && cache_->HasValue(i, j))
      return cache_->GetValue(i, j);
    if ((i == j) && (type & kCacheNorm) && cache_->HasNorm(i))
      return cache_->GetNorm(i);
    double value;
    DataSet<Object> *dataset = KernelImpl<Object>::Data();
    if ((type & kCacheFeature)) {
      typename F::Feature *fi, *fj;
      if (cache_->HasFeature(i)) {
        fi = cache_->GetFeature(i);
      } else {
        fi = function_->GetFeature((*dataset)[i]);
        cache_->SetFeature(i, fi);
      }
      if (cache_->HasFeature(j)) {
        fj = cache_->GetFeature(j);
      } else {
        fj = function_->GetFeature((*dataset)[j]);
        cache_->SetFeature(j, fj);
      }
      value = function_->Product(*fi, *fj);
    } else {
      value = (*function_)((*dataset)[i], (*dataset)[j]);
    }
    if (type & kCacheValue)
      cache_->SetValue(i, j, value);
    else if ((i == j) && (type & kCacheNorm))
      cache_->SetNorm(i, value);
    return value;
  }

  static BasicKernelImpl<F,C> *Read(istream &strm,
                                    const KernelReadOptions &opts);

  bool Write(ostream &strm, const KernelWriteOptions &opts) const;

  void ClearCache() { if (cache_) cache_->Clear(); }

  const F* GetFunction() { return function_; }

 private:
  BasicKernelImpl() : cache_(0), function_(0), normalize_(false) {}

 private:
  C *cache_;
  const F *function_;
  bool normalize_;

  // Identifies current file format version
  static const int32 kFileVersion = 2;
  // Identifies minimum compatible file format version
  static const int32 kMinFileVersion = 2;
};

template <class F, class C>
BasicKernelImpl<F,C> *BasicKernelImpl<F,C>::Read(
    istream &strm, const KernelReadOptions &opts) {

  BasicKernelImpl<F,C> *impl = new BasicKernelImpl<F,C>();
  impl->SetType(opts.type);
  KernelHeader hdr;
  if (!impl->ReadHeader(strm, opts, kMinFileVersion, &hdr))
    return 0;

  ReadType(strm, &impl->normalize_);
  impl->function_ = F::Read(strm, opts);
  impl->cache_ = C::Read(strm, opts);
  if (!impl->function_ || !impl->cache_) {
    LOG(ERROR) << "BasicKernel::Read: read failed: " << opts.source;
    return 0;
  }

  bool has_ds;
  ReadType(strm, &has_ds);
  if (has_ds) {
    DataSet<typename F::Object> *ds =
      DataSet<typename F::Object>::Read(strm, opts);
    impl->SetData(ds);
  }
  return impl;
}

template <class F, class C>
bool BasicKernelImpl<F,C>::Write(ostream &strm,
                                 const KernelWriteOptions &opts) const {
  KernelHeader hdr;
  KernelImpl<Object>::WriteHeader(strm, opts, kFileVersion, &hdr);
  if (!strm)
    return false;

  WriteType(strm, normalize_);
  function_->Write(strm, opts);
  cache_->Write(strm, opts);
  const DataSet<typename F::Object> *ds = this->Data();
  bool has_ds = (ds != 0);
  WriteType(strm, has_ds);
  if (has_ds)
     ds->Write(strm, opts);
  strm.flush();
  if (!strm)
    LOG(ERROR) << "BasicKernel::Write: write failed: " << opts.source;
  return strm;
}


template <class F, class C>
class BasicKernel : public Kernel<typename F::Object> {
 public:
  typedef F Function;
  typedef C Cache;
  typedef typename Function::Object Object;

  BasicKernel(const F* f, C* c, bool n) :
      impl_(new BasicKernelImpl<F,C>(f,c,n)) {}

  BasicKernel(const BasicKernel<F,C> &kernel) : impl_(kernel.impl_) {
    impl_->IncrRefCount();
  }

  virtual ~BasicKernel() { if(!impl_->DecrRefCount()) delete impl_;}

  BasicKernel<F,C> &operator=(const BasicKernel<F,C> &kernel) {
    if (this != &kernel) {
      if(!impl_->DecrRefCount()) delete impl_;
      kernel.impl_->IncrRefCount();
      impl_ = kernel.impl_;
    }
    return *this;
  }

  virtual double Eval(size_t i, size_t j) {
    return impl_->Eval(i,j);
  }

  virtual const string& Type() const { return impl_->Type(); }

  virtual BasicKernel<F,C> *Copy() const {
    impl_->IncrRefCount();
    return new BasicKernel<F,C>(impl_);
  }

  static BasicKernel<F,C> *Read(istream &strm, const KernelReadOptions &opts) {
    BasicKernelImpl<F,C>* impl = BasicKernelImpl<F,C>::Read(strm, opts);
    return impl ? new BasicKernel<F,C>(impl) : 0;
  }

  static BasicKernel<F,C> *Read(const string &filename) {
    if (!filename.empty()) {
      ifstream strm(filename.c_str());
      if (!strm) {
        LOG(ERROR) << "BasicKernel::Read: Can't open file: " << filename;
        return 0;
      }
      return Read(strm, KernelReadOptions(filename));
    } else {
      return Read(std::cin, KernelReadOptions("standard input"));
    }
  }

  virtual bool Write(ostream &strm, const KernelWriteOptions &opts) const {
    return impl_->Write(strm, opts);
  }

  virtual bool Write(const string &filename) const {
    if (!filename.empty()) {
      ofstream strm(filename.c_str());
      if (!strm) {
        LOG(ERROR) << "BasicKernel::Write: Can't open file: " << filename;
        return false;
      }
      return Write(strm, KernelWriteOptions(filename));
    } else {
      return Write(std::cout, KernelWriteOptions("standard output"));
    }
  }

  virtual void SetData(DataSet<Object>* ds) {
    impl_->SetData(ds);
    impl_->ClearCache();
  }

  virtual const DataSet<Object>* Data() const {
    return impl_->Data();
  }

  virtual DataSet<Object>* Data() {
    return impl_->Data();
  }

  virtual uint64 Properties(uint64 mask) const {
    return impl_->Properties(mask);
  }

  virtual void SetProperties(uint64 props, uint64 mask) {
    impl_->SetProperties(props, mask);
  }

  const F* GetFunction() { return impl_->GetFunction(); }

 protected:
  BasicKernelImpl<F,C> *Impl() {return impl_;}
  explicit BasicKernel(BasicKernelImpl<F,C> *impl) : impl_(impl) {}

 private:
  BasicKernelImpl<F,C> *impl_;
};


}  // namespace kernel

#endif  // KERNEL_LIB_BASIC_KERNEL_H__
