
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Author: allauzen@google.com (Cyril Allauzen)

#ifndef KERNEL_LIB_BLAS_H_
#define KERNEL_LIB_BLAS_H_

#include "kernel/lib/compat.h"



namespace kernel {

using lapack::integer;
using lapack::real;
using lapack::doublereal;

template <typename T>
struct BlasVectorInfo {
  integer dimension;
  integer increment;
  T *array;
};

template <typename T>
struct BlasMatrixInfo {
  char transpose;
  integer leading_dimension;
  integer other_dimension;
  const T *array;
};

template <typename T>
struct MutableBlasMatrixInfo {
  char transpose;
  integer leading_dimension;
  integer other_dimension;
  T *array;
};


/////////////////////////////////////////////////////////////////////////
//
// Lapack
//
// Interface functions behave exactly as documented in Lapack. The lapack
// documentation is included below for convenience.
//
/////////////////////////////////////////////////////////////////////////

/*
  Purpose
  =======

  DGESV computes the solution to a real system of linear equations
     A * X = B,
  where A is an N-by-N matrix and X and B are N-by-NRHS matrices.
*/
template <typename T>
struct Gesv;

template <>
struct Gesv<double> {
  void operator()(integer *n, integer *nrhs, doublereal *a,
                  integer *lda, integer *ipiv, doublereal *b, 
                  integer *ldb, integer *info) {
    ::dgesv_(n, nrhs, a, lda, ipiv, b, ldb, info);
  }
};

template <>
struct Gesv<float> {
  void operator()(integer *n, integer *nrhs, real *a,
                  integer *lda, integer *ipiv, real *b, 
                  integer *ldb, integer *info) {
    ::sgesv_(n, nrhs, a, lda, ipiv, b, ldb, info);
  }
};

/*
  Purpose
  =======

  DPOSV computes the solution to a real system of linear equations
     A * X = B,
  where A is an N-by-N symmetric positive definite matrix and X and B
  are N-by-NRHS matrices.

 =====================================================================
*/
template <typename T>
struct Posv;

template <>
struct Posv<double> {
  void operator()(char *uplo, integer *n, integer *nrhs, doublereal *a,
                  integer *lda, doublereal *b, integer *ldb, integer
                  *info) {
    ::dposv_(uplo, n, nrhs, a, lda,
             b, ldb, info);
  }
};

template <>
struct Posv<float> {
  void operator()(char *uplo, integer *n, integer *nrhs, real *a,
                  integer *lda, real *b, integer *ldb, integer *info) {
    ::sposv_(uplo, n, nrhs, a, lda, b,
             ldb, info);
  }
};


//
// _SYEV:
//
/*
  Purpose
  =======

  DSYEV computes all eigenvalues and, optionally, eigenvectors of a
  real symmetric matrix A.

  =====================================================================
*/
template <typename T>
struct Syev;

template <>
struct Syev<double> {
  void operator()(char *jobz, char *uplo, integer *n, doublereal *a,
                  integer *lda, doublereal *w, doublereal *work,
                  integer *lwork, integer *info) {
    ::dsyev_(jobz, uplo, n, a, lda, w, work, lwork, info);
  }
};

template <>
struct Syev<float> {
  void operator()(char *jobz, char *uplo, integer *n, real *a,
                  integer *lda, real *w, real *work,
                  integer *lwork, integer *info) {
    ::ssyev_(jobz, uplo, n, a, lda, w, work, lwork, info);
  }
};

//
// _POTRI:
//
/*
  Purpose
  =======

  DPOTRI computes the inverse of a real symmetric positive definite
  matrix A using the Cholesky factorization A = U**T*U or A = L*L**T
  computed by DPOTRF.

  =====================================================================
*/
template <typename T>
struct Potri;

template<>
struct Potri<double> {
  void operator()(char *uplo, integer *n, doublereal *a, integer *lda,
                  integer *info) {
    ::dpotri_(uplo, n, a, lda, info);
  }
};

template<>
struct Potri<float> {
  void operator()(char *uplo, integer *n, real *a, integer *lda,
                  integer *info) {
    ::spotri_(uplo, n, a, lda, info);
  }
};

//
// _POTRF:
//
/*
  Purpose
  =======

  DPOTRF computes the Cholesky factorization of a real symmetric
  positive definite matrix A.

  The factorization has the form
     A = U**T * U,  if UPLO = 'U', or
     A = L  * L**T,  if UPLO = 'L',
  where U is an upper triangular matrix and L is lower triangular.

 =====================================================================
*/
template <typename T>
struct Potrf;

template<>
struct Potrf<double> {
  void operator()(char *uplo, integer *n, doublereal *a, integer *lda,
                  integer *info) {
    ::dpotrf_(uplo, n, a, lda, info);
  }
};

template<>
struct Potrf<float> {
  void operator()(char *uplo, integer *n, real *a, integer *lda,
                  integer *info) {
    ::spotrf_(uplo, n, a, lda, info);
  }
};

//
// _GETRI:
//
/*
  Purpose
  =======

  DGETRI computes the inverse of a matrix using the LU factorization
  computed by DGETRF.

  This method inverts U and then computes inv(A) by solving the system
  inv(A)*L = inv(U) for inv(A).

 =====================================================================
*/
template <typename T>
struct Getri;

template<>
struct Getri<double> {
  void operator()(integer *n, doublereal *a, integer *lda, integer *ipiv,
                  doublereal *work, integer *lwork, integer *info) {
    ::dgetri_(n, a, lda, ipiv, work, lwork, info);
  }
};

template<>
struct Getri<float> {
  void operator()(integer *n, real *a, integer *lda, integer *ipiv,
                  real *work, integer *lwork, integer *info) {
    ::sgetri_(n, a, lda, ipiv, work, lwork, info);
  }
};


//
// _GETRF:
//
/*
  Purpose
  =======

  DGETRF computes an LU factorization of a general M-by-N matrix A
  using partial pivoting with row interchanges.

  The factorization has the form
     A = P * L * U
  where P is a permutation matrix, L is lower triangular with unit
  diagonal elements (lower trapezoidal if m > n), and U is upper
  triangular (upper trapezoidal if m < n).

  =====================================================================
*/
template <typename T>
struct Getrf;

template<>
struct Getrf<double> {
  void operator()(integer *m, integer *n, doublereal *a, integer *lda,
                  integer *ipiv, integer *info) {
    ::dgetrf_(m, n, a, lda, ipiv, info);
  }
};

template<>
struct Getrf<float> {
  void operator()(integer *m, integer *n, real *a, integer *lda,
                  integer *ipiv, integer *info) {
    ::sgetrf_(m, n, a, lda, ipiv, info);
  }
};

/////////////////////////////////////////////////////////////////////////
//
// LEVEL 3 Blas
//
/////////////////////////////////////////////////////////////////////////

//
// _GEMM:  C = alpha * A * B + beta * C
//

// Generic undefined wrapper
template <typename T>
struct Gemm;

// Specialization for double
template <>
struct Gemm<double> {
  void operator()(char *transa, char *transb, integer *m, integer *n,
                  integer *k, doublereal *alpha, const doublereal *a,
                  integer *lda, const doublereal *b, integer *ldb,
                  doublereal *beta, doublereal *c, integer *ldc) {
    ::dgemm_(transa, transb, m, n, k, alpha, const_cast<doublereal*>(a), lda,
             const_cast<doublereal*>(b), ldb, beta, c, ldc);
  }
};

// Specialization for float
template<>
struct Gemm<float> {
  void operator()(char *transa, char *transb, integer *m, integer *n,
                  integer *k, real *alpha, const real *a,
                  integer *lda, const real *b, integer *ldb,
                  real *beta, real *c, integer *ldc) {
    ::sgemm_(transa, transb, m, n, k, alpha, const_cast<real*>(a), lda,
             const_cast<real*>(b), ldb, beta, c, ldc);
  }
};

//
// _SYMM:  C = alpha * A * B + beta * C or
//         C = alpha * B * A + beta * C
//         with A^T = A
//

// Generic undefined wrapper
template <typename T>
struct Symm;

// Specialization for double
template <>
struct Symm<double> {
  void operator()(char *side, char *uplo, integer *m, integer *n,
                  doublereal *alpha, doublereal *a, integer *lda,
                  doublereal *b, integer *ldb, doublereal *beta,
                  doublereal *c, integer *ldc) {
    ::dsymm_(side, uplo, m, n, alpha, a, lda, b, ldb, beta, c, ldc);
  }
};

// Specialization for float
template <>
struct Symm<float> {
  void operator()(char *side, char *uplo, integer *m, integer *n,
                  real *alpha, real *a, integer *lda,
                  real *b, integer *ldb, real *beta,
                  real *c, integer *ldc) {
    ::ssymm_(side, uplo, m, n, alpha, a, lda, b, ldb, beta, c, ldc);
  }
};

//
// _SYRK: C = alpha * A * A^T + beta *C or
//        C = alpha * A^T * A + beta *C
//        with C^T = C
//

// Generic undefined wrapper
template <typename T>
struct Syrk;

// Specialization for double
template <>
struct Syrk<double> {
  void operator()(char *uplo, char *trans, integer *n, integer *k,
                  doublereal *alpha, doublereal *a, integer *lda,
                  doublereal *beta, doublereal *c, integer *ldc) {
    ::dsyrk_(uplo, trans, n, k, alpha, a, lda, beta, c, ldc);
  }
};

// Specialization for double
template <>
struct Syrk<float> {
  void operator()(char *uplo, char *trans, integer *n, integer *k,
                  real *alpha, real *a, integer *lda,
                  real *beta, real *c, integer *ldc) {
    ::ssyrk_(uplo, trans, n, k, alpha, a, lda, beta, c, ldc);
  }
};


/////////////////////////////////////////////////////////////////////////
//
// LEVEL 2 Blas
//
/////////////////////////////////////////////////////////////////////////

//
// _GEMV:  y = alpha * A * x + beta * y
//

// Generic undefined wrapper
template <typename T>
struct Gemv;

// Specialization for double
template <>
struct Gemv<double> {
  void operator()(char *trans, integer *m, integer *n, doublereal *alpha,
                  const doublereal *a, integer *lda, doublereal *x,
                  integer *incx, doublereal *beta, doublereal *y,
                  integer *incy) {
    ::dgemv_(trans, m, n, alpha, const_cast<doublereal*>(a), lda, x, incx,
             beta, y, incy);
  }
};

// Specialization for float
template <>
struct Gemv<float> {
  void operator()(char *trans, integer *m, integer *n, real *alpha,
                  const real *a, integer *lda, real *x, integer *incx,
                  real *beta, real *y, integer *incy) {
    ::sgemv_(trans, m, n, alpha, const_cast<real*>(a), lda, x, incx, beta,
             y, incy);
  }
};


}  // namespace kernel

#endif  // KERNEL_LIB_BLAS_H_
