// object.h
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Author: allauzen@cs.nyu.edu (Cyril Allauzen)
//
// \file
// Object representation.

#ifndef KERNEL_LIB_OBJECT_H__
#define KERNEL_LIB_OBJECT_H__

#include <fst/fst.h>
#include <fst/util.h>
#include <sstream>

namespace kernel {

struct KernelReadOptions;
struct KernelWriteOptions;

using fst::Fst;
using fst::FstReadOptions;
using fst::FstWriteOptions;
using fst::SplitToVector;
using fst::StrToWeight;
using fst::istream;
using fst::ostream;
using fst::istringstream;

//////////////////////////////////////////////////////////////////////
//
// Functors handling the manipulations of objects
//
//////////////////////////////////////////////////////////////////////

template <class O>
struct ObjectWrite {
  bool operator()(ostream &strm, const O &object,
                  const KernelWriteOptions &opts = KernelWriteOptions()) {
    return object.Write(strm);  // TODO(allauzen): add opts or source?
  }
};

template <class O>
struct ObjectRead {
  O *operator()(istream &strm,
                const KernelReadOptions &opts = KernelReadOptions()) {
    return O::Read(strm);  // TODO(allauzen): add opts or source?
  }
};

template <class O>
struct ObjectReadText {
  O *operator()(const string &txt, const string &src, size_t nline) {
    LOG(ERROR) << "Unsupported"; // TODO(allauzen): clean this mess!
    return 0;
  }
};

template <class O>
struct ObjectType {
  static const string type;
};

template <class O>
struct ObjectCopy {
  O *operator()(const O&object) {
    return new O(object);
  }
};


//////////////////////////////////////////////////////////////////////
//
// Specializations for the "Fst<Arc>" object type
//
//////////////////////////////////////////////////////////////////////

template <>
template <class Arc>
struct ObjectWrite< Fst<Arc> > {
  bool operator()(ostream &strm, const Fst<Arc> &fst,
                  const KernelWriteOptions &opts = KernelWriteOptions()) {
    return fst.Write(strm, FstWriteOptions());  // TODO(allauzen): opts.source
  }
};


template <>
template <class Arc>
struct ObjectRead< Fst<Arc> > {
  Fst<Arc>* operator()(istream &strm,
                       const KernelReadOptions &opts = KernelReadOptions()) {
    // TODO(allauzen): opts.source
    return Fst<Arc>::Read(strm, FstReadOptions());
  }
};

template <>
template <class Arc>
struct ObjectReadText< Fst<Arc> > {
  Fst<Arc>* operator()(const string &text, const string &src, size_t nline) {
    ifstream is(text.c_str());
    Fst<Arc> *fst = Fst<Arc>::Read(is, FstReadOptions(text));
    return fst;
  }
};

template <>
template <class Arc>
struct ObjectCopy< Fst<Arc> > {
  Fst<Arc>* operator()(const Fst<Arc> &fst) {
    return fst.Copy();
  }
};

//////////////////////////////////////////////////////////////////////
//
// Specializations for the "SparseVector<T, I>" object type
//
//////////////////////////////////////////////////////////////////////

// This should be added to io-utilities.h and the functions there
// should be be rewritten to re-use to use this
template <class V>
void StringToVector(const string &str, V *vec, const string &src,
                    size_t nline) {
  char *c_str = new char[str.size() + 1];
  str.copy(c_str, str.size());
  c_str[str.size()] = 0;
  vec->Clear();
  vector<char *> entries;
  SplitToVector(c_str, " \t", &entries, true);
  vec->Reserve(entries.size());
  vector<char *> pair;
  for (size_t i = 0; i < entries.size(); ++i) {
    pair.clear();
    SplitToVector(entries[i], ":", &pair, false);
    if ((i == 0) && (pair.size() == 1))
        continue;
    CHECK_EQ(pair.size(), 2);
    vec->AddElement(
        StrToWeight<typename V::IndexType>(string(pair[0]), src, nline),
        StrToWeight<typename V::ElementType>(string(pair[1]), src, nline));
  }  // TODO: all this mess need to be fixed
  delete[] c_str;
}


// Forward declaration of sparse vector
template <class T, class I, class D>
class SparseVector;


template <>
template <class T, class I, class D>
struct ObjectReadText<SparseVector<T, I, D> > {
  SparseVector<T, I, D> *operator()(const string &text,
                                    const string &src,
                                    size_t nline) {
    SparseVector<T, I, D> *sv = new SparseVector<T, I, D>;
    StringToVector(text, sv, src, nline);
    return sv;
  }
};

}  // namespace kernel

#endif  // KERNEL_LIB_OBJECT_H__
