// sum-kernel.h
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Author: allauzen@cs.nyu.edu (Cyril Allauzen)
//
// \file
// Sum kernel representation.

#ifndef KERNEL_LIB_SUM_KERNEL_H__
#define KERNEL_LIB_SUM_KERNEL_H__


#include "kernel/lib/kernel.h"

namespace kernel {


template <class O>
class SumKernelImpl : public KernelImpl<O> {
 public:
  typedef O Object;

  SumKernelImpl(const Kernel<O> &k1, const Kernel<O> &k2,
                double a1, double a2) {
    kernels_.push_back(k1.Copy());
    kernels_.push_back(k2.Copy());
    alphas_.push_back(a1);
    alphas_.push_back(a2);
    Init();
  }


  SumKernelImpl(const vector<const Kernel<O> *> kernels,
                const vector<double> alphas) : alphas_(alphas) {
    if (kernels.size() != alphas.size())
      LOG(FATAL) << "SumKernel: # kernels is not equal to # coefficient";
    if (kernels.size() == 0)
      LOG(FATAL) << "SumKernel: need to specify at least one input kernel";
    for (size_t i = 0; i < kernels.size(); ++i)
      kernels_.push_back(kernels_[i]->Copy());
    Init();
  }

  ~SumKernelImpl() {
    for (size_t i = 0; i < kernels_.size(); ++i)
      delete kernels_[i];
  }

  double Eval(size_t i, size_t j) {
    double result = 0.0;
    for (size_t k = 0; k < alphas_.size(); ++k)
      result += alphas_[k] * kernels_[k]->Eval(i, j);
    return result;
  }

  void SetData(DataSet<O> *ds) {
    for (size_t i = 0; i < kernels_.size(); ++i)
      kernels_[i]->SetData(ds);
    KernelImpl<O>::SetData(ds);
  }

  static SumKernelImpl<O> *Read(istream &strm, const KernelReadOptions &opts) {
    SumKernelImpl<O> *impl = new SumKernelImpl<O>();
    KernelHeader hdr;
    if (!impl->ReadHeader(strm, opts, kMinFileVersion, &hdr))
        return 0;
    if (hdr.Version() == 1) {
      impl->alphas_.resize(2);
      ReadType(strm, &(impl->alphas_[0]));
      ReadType(strm, &(impl->alphas_[1]));
      if (!strm) {
        LOG(ERROR) << "SumKernel::Read: read failed: " << opts.source;
        return 0;
      }
      impl->kernels_.resize(2);
      impl->kernels_[0] = down_cast<Kernel<O>*>(
          Kernel<O>::Read(strm, KernelReadOptions(opts.source)));
      if (!impl->kernels_[0]) {
        LOG(ERROR) << "SumKernel::Read: read failed: " << opts.source;
        return 0;
      }
      impl->kernels_[1] = down_cast<Kernel<O>*>(
          Kernel<O>::Read(strm, KernelReadOptions(opts.source)));
      if (!impl->kernels_[1]) {
        LOG(ERROR) << "SumKernel::Read: read failed: " << opts.source;
        return 0;
      }
      impl->InitDataSet();
      return impl;
    }

    ReadType(strm, &(impl->alphas_));
    if (!strm) {
      LOG(ERROR) << "SumKernel::Read: read failed: " << opts.source;
      return 0;
    }
    int64 num;
    ReadType(strm, &num);
    for (int64 i = 0; i < num; ++i) {
      Kernel<O> *kernel = down_cast<Kernel<O>*>(
          Kernel<O>::Read(strm, KernelReadOptions(opts.source)));
      if (!kernel) {
        LOG(ERROR) << "SumKernel::Read: read failed: " << opts.source;
        return 0;
      }
      impl->kernels_.push_back(kernel);
    }
    impl->InitDataSet();
    return impl;
  }

  bool Write(ostream &strm, const KernelWriteOptions &opts) const {
    KernelHeader hdr;
    KernelImpl<O>::WriteHeader(strm, opts, kFileVersion, &hdr);
    if (!strm)
      return false;
    WriteType(strm, alphas_);
    int64 num = kernels_.size();
    WriteType(strm, num);
    for (size_t i = 0; i < kernels_.size(); ++i)
      kernels_[i]->Write(strm, KernelWriteOptions(opts.source));
    if (!strm)
      LOG(ERROR) << "SumKernel::Write: write failed: " << opts.source;
    return strm;
  }

 private:
  void Init() {
    KernelImpl<O>::SetType("sum");
    InitProperties();
    InitDataSet();
  }

  void InitProperties() {
        uint64 props = kernels_[0]->Properties();
    double alpha = alphas_[0];
    for (size_t i = 1; i < kernels_.size(); ++i) {
      props = SumProperties(props, kernels_[i]->Properties(),
                            alpha, alphas_[i]);
      alpha += alphas_[i];
    }
    KernelImpl<O>::SetProperties(props);
  }

  void InitDataSet() {
    size_t i = 0;
    for (; i < alphas_.size(); ++i)
      if (alphas_[i] != 0) break;
    DataSet<O> *ds = kernels_[i]->Data();
    for (++i; i < alphas_.size(); ++i)
      if ((alphas_[i] != 0) && (kernels_[i]->Data() != ds)) {
        ds = 0;
        break;
      }
    KernelImpl<O>::SetData(ds);
  }

  SumKernelImpl() { KernelImpl<O>::SetType("sum"); }

 private:
  vector<Kernel<O>*> kernels_;
  vector<double> alphas_;

  // Identifies current file format version
  static const int32 kFileVersion = 2;
  // Identifies minimum compatible file format version
  static const int32 kMinFileVersion = 1;
};


template <class O>
class SumKernel :public Kernel<O> {
 public:
  typedef O Object;

  SumKernel(const Kernel<O> &k1, const Kernel<O> &k2,
            double a1 = 1.0F , double a2 = 1.0F) :
      impl_(new SumKernelImpl<O>(k1, k2, a1, a2)) {}

  SumKernel(const vector<const Kernel<O>*> &kernels, vector<double> alphas)
      : impl_(new SumKernelImpl<O>(kernels, alphas)) {}

  SumKernel(const SumKernel<O> &kernel) : impl_(kernel.impl_) {
    impl_->IncrRefCount();
  }

  virtual ~SumKernel() { if(!impl_->DecrRefCount()) delete impl_;}

  SumKernel<O> &operator=(const SumKernel<O> &kernel) {
    if (this != &kernel) {
      if(!impl_->DecrRefCount()) delete impl_;
      kernel.impl_->IncrRefCount();
      impl_ = kernel.impl_;
    }
    return *this;
  }

  virtual double Eval(size_t i, size_t j) {
    return impl_->Eval(i,j);
  }

  virtual const string& Type() const { return impl_->Type(); }

  virtual SumKernel<O> *Copy() const {
    impl_->IncrRefCount();
    return new SumKernel<O>(impl_);
  }

  // Read and Write
  static SumKernel<O> *Read(istream &strm,
                                   const KernelReadOptions &opts) {
    SumKernelImpl<O> *impl = SumKernelImpl<O>::Read(strm, opts);
    return impl ? new SumKernel(impl) : 0;
  }

  static SumKernel<O> *Read(const string &filename) {
    if (!filename.empty()) {
      ifstream strm(filename.c_str());
      if (!strm) {
        LOG(ERROR) << "SumKernel::Read: Can't open file: " << filename;
        return 0;
      }
      return Read(strm, KernelReadOptions(filename));
    } else {
      return Read(std::cin, KernelReadOptions("standard input"));
    }
  }

  virtual bool Write(ostream &strm, const KernelWriteOptions &opts) const {
    impl_->Write(strm, opts);
    return strm;
  }

  virtual bool Write(const string &filename) const {
    if (!filename.empty()) {
      ofstream strm(filename.c_str());
      if (!strm) {
        LOG(ERROR) << "SumKernel::Write: Can't open file: " << filename;
        return false;
      }
      return Write(strm, KernelWriteOptions(filename));
    } else {
      return Write(std::cout, KernelWriteOptions("standard output"));
    }
  }

  virtual void SetData(DataSet<O>* ds) {
    impl_->SetData(ds);
  }

  virtual const DataSet<O>* Data() const {
    return impl_->Data();
  }

  virtual DataSet<Object>* Data() {
    return impl_->Data();
  }

  virtual uint64 Properties(uint64 mask) const {
    return impl_->Properties(mask);
  }

  virtual void SetProperties(uint64 props, uint64 mask) {
    impl_->SetProperties(props, mask);
  }

 protected:
  explicit SumKernel(SumKernelImpl<O> *impl) : impl_(impl) {}

 private:
  SumKernelImpl<O> *impl_;
};

}  // namespace kernel

#endif  // KERNEL_LIB_SUM_KERNEL_H__
